import os
import pandas as pd
import numpy as np
import seaborn as sns
import anndata
from scipy import optimize, interpolate
from scipy.stats import f as f_distribution
from scipy.stats import t as t_distribution

doses= np.array([0.0,
   0.03,
   0.3,
   1.0,
   3.0,
   10.0,
   30.0,
   100.0,
   300.0,
   1000.0,
   10000.0])


#### These line shave been copied from Florian Bayers code from Gitlab ######

def logistic_model(x, ec50, slope, front, back):
    """
    logisitc_model(x, ec50, slope, top, bottom)

    Logistic model to fit the drug response data.

    Parameters
    ----------
    x: array-like
        Drug concentrations in log space
    ec50: float
        Inflection point in log space of the sigmoid curve
    slope: float
        slope of the transition between front and back asymptote
    front: float
        front (first) asymptote
    back: float
        back (second) asymptote

    Returns
    -------
    y: array-like
        Drug response
    """
    # predict y with given parameters using the standard model
    return (front - back) / (1 + 10 ** (slope * (x - ec50))) + back


def get_s0(fc_lim, alpha):
    """
    Calculates the s0 value given a log2 fold change limit and an alpha value.
    This is based on the SAM test analysis.
    """
    log_alpha = np.log10(alpha)
    k = fc_lim / (1/2 * log_alpha + 1/10)
    s0 = abs(np.log10(2**k))
    return s0

def map_fc_to_p_samcutoff(x, alpha, s0, deg_f):
    """
    This is the SAM test for a fold change

    x : log2 fold change value

    alpha : significance threshold
    s0 : fudge factor for variance based on SAM-Test idea
    deg_f : degrees of freedom

    Adapted from:
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/10.1002/pmic.201600132
    R Code in the supplement
    """
    # Get the t-value for given significance threshold
    ta = t_distribution.ppf(1 - alpha, deg_f)

    # Mask
    pos = x > (ta * s0)
    neg = x < (-ta * s0)

    # Values
    x_pos = x[pos]
    x_neg = x[neg]
    x_none = x[(~pos) & (~neg)]

    d_pos = x_pos / ta - s0
    d_pos = s0 / d_pos
    d_pos = ta * (1 + d_pos)

    d_neg = x_neg / (-ta) - s0
    d_neg = s0 / d_neg
    d_neg = ta * (1 + d_neg)

    # cumulative density function (cdf) of the Student t distribution
    # Log survival function to have more accurate p value calculations:  - np.log10[(1 - t_dist.cdf(t_val, deg_f)) * 2]
    y_pos = - t_distribution.logsf(d_pos, deg_f) * np.log10(np.e) - np.log10(2)
    y_neg = - t_distribution.logsf(d_neg, deg_f) * np.log10(np.e) - np.log10(2)

    # Combine array
    y_none = np.full(shape=len(x_none), fill_value=np.inf)
    y = pd.concat([pd.Series(y_neg, index=x_neg.index),
                   pd.Series(y_none, index=x_none.index),
                   pd.Series(y_pos, index=x_pos.index)])
    return y[x.index]



def sam_curve(x, deg_f=10-2, fc_lim=np.log2(1.3), alpha=0.015):
    s0=get_s0(fc_lim, alpha)
    ta = t_distribution.ppf(1 - alpha, deg_f)
    d_pos = x / ta - s0
    d_pos = s0 / d_pos
    d_pos = ta * (1 + d_pos)
    y = - t_distribution.logsf(d_pos, deg_f) * np.log10(np.e) - np.log10(2)
    return(y)


def add_fold_change(DF, doses=doses):
    front_predicted=[]
    back_predicted=[]
    fold_change_predicted=[]

    Curve_fits=DF[['Log EC50', 'Curve Slope', 'Curve Front', 'Curve Back']].values

    for i in range(Curve_fits.shape[0]):

        ec50, slope, front, back=Curve_fits[i,:]

        front_predicted.append(front)
        back_predicted.append(logistic_model(np.log10(doses[-1]/10**9), ec50, slope, front, back))
        fold_change_predicted.append(back_predicted[-1]/front)

    DF['Front predicted']=front_predicted
    DF['Back predicted']=back_predicted
    DF['Fold_change_predicted']=fold_change_predicted
    DF['log(Fold_change_predicted)']=np.log(fold_change_predicted)
    DF['0']=0
    
    curve_regualtion_old=['-'*20]*len(DF)
    
    for i in range(len(DF)):
        cr=DF['Curve Regulation'].iloc[i]
        if str(cr)=='nan':
            curve_regualtion_old[i]='Unknown'
        elif str(cr)=='not':
            curve_regualtion_old[i]='Not regulated'
        elif str(cr)=='up':
            curve_regualtion_old[i]='Up-regulated'   
        elif str(cr)=='down':
            curve_regualtion_old[i]='Down-regulated' 
        else:
            curve_regualtion_old[i]=cr
    DF['Old_curve_regulation']=curve_regualtion_old
    return(DF)



def calculate_new_regulation(df, alpha=0.015, fc_lim=0.39, new_not_regualted=True):

    Fold_change_key='Fold_change_predicted'
    Q_value_key='Curve Log P_Value'

    ############## This is a reduced version of Flo's "define_regulated_curves"    ##################################

    fold_change = np.log2(df[Fold_change_key].copy())
    s0 = get_s0(fc_lim, alpha)
    p_value_cutoff = map_fc_to_p_samcutoff(fold_change, alpha=alpha, s0=s0, deg_f=10-2)
    p_mask = df[Q_value_key] >= p_value_cutoff
    down_mask = fold_change < 0
    up_mask = fold_change > 0

    # Not regulated curves require: not too much noise, not too much deviation to control
    # TODO: Do we need a p value cut off ? currently up and down regulation will overwrite not,
    # also p value might be strong but no effect
    rmse_max = 0.1
    
    if new_not_regualted:
        not_regulated_mask = (df['Curve Simple Model RMSE'] < rmse_max) & \
                             (np.log2(df[Fold_change_key])).between(-fc_lim, fc_lim)
    
    else:
        not_regulated_mask = (df['Curve Simple Model RMSE'] < rmse_max) & \
                             (np.log2(df['Curve Simple Model'])).between(-abs(fc_lim)/2, abs(fc_lim)/2)


    # Add labels
    df['New_curve_regulation'] = 'Unknown'
    df.loc[not_regulated_mask, 'New_curve_regulation'] = 'Not regulated'
    df.loc[p_mask & up_mask, 'New_curve_regulation'] = 'Up-regulated'
    df.loc[p_mask & down_mask, 'New_curve_regulation'] = 'Down-regulated'
    
    return(df)

def pick_curve(dfsp, pick_criterion='Andromeda Score'):
    
    if pick_criterion=='Percolator Score':
        index_of_max=np.argmax(dfsp['Percolator Score'])
    elif pick_criterion=='Andromeda Score':
        index_of_max=np.argmax(dfsp['Andromeda Score'])
    elif pick_criterion=='Total S/N':
        index_of_max=np.argmax(dfsp['Andromeda Score'])
    elif pick_criterion=='Curve P_Value':
        index_of_max=np.argmin(dfsp['Curve P_Value'])
    else:
        print('Error: pick_criterion is not in [Percolator Score, Andromeda Score, Total S/N, Curve P_value]')
    
    dfss=dfsp.iloc[[index_of_max]]
        
    return(dfss)

def select_curves(df, pick_criterion='Andromeda Score'):
    df['Modified sequence']=df['Modified sequence'].astype('category')

    peptides=[a for a in df['Modified sequence'].cat.categories]
    
    frames=[]

    for peptide in peptides:

        dfp=df[df['Modified sequence']==peptide].copy()

        dfs=pick_curve(dfp, pick_criterion=pick_criterion)

        frames.append(dfs)

    Df=pd.concat(frames)
    Df.index=np.arange(len(Df))
    
    return(Df)


def create_dicts(Data2):
    drugs=[a for a in Data2.keys()]
    
    Pep_dict={}
    Prot_dict={}
    Gene_dict={}
    
    
    peptides=[]

    l=0
    for i in range(len(drugs)):
        drug=drugs[i]
        df=Data2[drug]

        for j in range(len(df)):
            peptide=df['Modified sequence'][j]

            if not peptide in peptides:
                Pep_dict[peptide]=l
                Prot_dict[peptide]=df['Proteins'][j]
                Gene_dict[peptide]=df['Gene Names'][j]
                l+=1
                peptides=[a for a in Pep_dict.keys()]
    return(Pep_dict, Prot_dict, Gene_dict)

def create_regulation_matrix(Data2):
    Pep_dict, Prot_dict, Gene_dict=create_dicts(Data2)
    drugs=[a for a in Data2.keys()]
    n0=len(drugs)
    n1=len(Pep_dict.keys())

    Z=np.full((n1,n0), 'Empty', dtype='<U50')

    for i in range(len(drugs)):

        drug=drugs[i]
        df=Data2[drug]

        for j in range(len(df)):
            Z[Pep_dict[df['Modified sequence'][j]],i]=df['New_curve_regulation'][j]

    matrix=anndata.AnnData(X=Z,  dtype=str)
    matrix.var['Drug']=drugs
    matrix.obs['Modified sequence']=[a for a in Pep_dict.keys()]
    matrix.obs['Proteins']=[Prot_dict[a] for a in Pep_dict.keys()]
    matrix.obs['Gene Names']=[Gene_dict[a] for a in Pep_dict.keys()]

    return(matrix)

def add_key_to_regulation_matrix(Data2, matrix, key):
    
    drugs=[a for a in Data2.keys()]
    
    n0=len(drugs)
    n1=len(matrix)
    
    Pep_dict={}
    for i in range(n1):
        Pep_dict[matrix.obs['Modified sequence'].iloc[i]]=i


    Z=np.full((n1,n0), '-', dtype='<U50')
    
    for i in range(n0):
        drug=drugs[i]
        df=Data2[drug].copy()

        for j in range(len(df)):
            Z[Pep_dict[df['Modified sequence'][j]],i]=df[key][j]

    matrix.layers[key]=Z

    return(matrix)


def recalculate_regulation(Data, fc_lim=0.55, alpha=0.015, pick_criterion='Andromeda Score', new_not_regualted=True):
    Data2={}

    drugs=[a for a in Data.keys()]

    for drug in drugs:

        df=Data[drug].copy()
        df=add_fold_change(df, doses)
        df=calculate_new_regulation(df, alpha=alpha, fc_lim=fc_lim, new_not_regualted=new_not_regualted)
        Df=select_curves(df, pick_criterion=pick_criterion)

        Data2[drug]=Df
    
    return(Data2)


def display_change_in_regualtion_classification(Data2):
    frames=[]
    for drug in Data2.keys():
        df=Data2[drug]

        before=[]
        after=[]
        peptide=[]

        for i in range(len(df)):
            before.append(df['Old_curve_regulation'][i])
            after.append(df['New_curve_regulation'][i])
            peptide.append(df['Modified sequence'][i])

        dfn=pd.DataFrame({'Regulation_prior':before, 'Regulation_posterior':after, 'Drug':drug, 'Peptide':peptide})
        frames.append(dfn)

    Df=pd.concat(frames)
    Df.index=np.arange(len(Df))

    Z=np.zeros((4,4))

    regs=['Unknown','Not regulated','Down-regulated','Up-regulated']
    
    old_reg=Df['Regulation_prior'].astype('category').cat.categories
    new_reg=Df['Regulation_posterior'].astype('category').cat.categories
    
    for i in old_reg:
        if not i in regs:
            print(f'Error: {i} is not in {regs} (old regulation)')
    for i in new_reg:
        if not i in regs:
            print(f'Error: {i} is not in {regs} (new regulation)')


    for i in range(4):
        e0=regs[i]
        for j in range(4):
            e1=regs[j]
            Dfn=Df[Df['Regulation_prior']==e0]
            Z[i,j]=len(Dfn[Dfn['Regulation_posterior']==e1])
    Dfz=pd.DataFrame(Z, columns = ['After: '+a for a in regs])
    Dfz.index=['Before: '+a for a in regs]
    display(Dfz)
    return(Df,Dfz)


def add_curve_fit_results(matrix, Data2):

    for key in ['log(Fold_change_predicted)', 'Total S/N', 'pEC50', 'Log EC50 Error']:
        matrix=add_key_to_regulation_matrix(Data2, matrix, key)

    # When trying to store/load the file with 'Total S/N' as obsm-key this returns an error due to the backslash
    matrix.layers['Total SN']=matrix.layers['Total S/N'].copy()
    del matrix.layers['Total S/N']
    
    return(matrix)


def extract_from_tomls(toml_path, df_path):
    tomls=[a for a in os.listdir(toml_path)]

    Data={}
    for topas in tomls:
        drug=topas[:-5]
        
        A=pd.read_csv(f'{df_path}/{drug}.txt', sep='\t', low_memory=False)
        Ts=A.iloc[np.where([str(a)!='nan' for a in A['Protein Names']])]

        Data[drug]=Ts
    return(Data)


def extract_and_refine_regualtion(toml_path, df_path, fc_lim=0.55, alpha=0.015, pick_criterion='Andromeda Score'):
    
    # Takes about 20 min, and you can safely ignore the DtypeWarning's
    Data=extract_from_tomls(toml_path, df_path)

    # Takes about 90 min to run
    Data2=recalculate_regulation(Data, fc_lim, alpha, pick_criterion, new_not_regualted=False)

    Df,Dfz=display_change_in_regualtion_classification(Data2)

    matrix=create_regulation_matrix(Data2)

    matrix=add_curve_fit_results(matrix, Data2)
    
    return(matrix, Data2, Df, Dfz)


import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import matplotlib.pyplot as plt
import scipy


import os
import sys
sys.path.append(os.getcwd().rsplit('/', 1)[0]+'/Functions')

import ALS as als
import Neural_network_RS as nn



def partition_matrix(matrix):
    n0,n1=matrix.shape
    
    S={'Unknown':[], 'Empty':[], 'Not regulated':[], 'Up-regulated':[], 'Down-regulated':[]}

    for i in range(n0):
        for j in range(n1):
            if matrix.X[i,j] in S.keys():
                S[matrix.X[i,j]].append((i,j))
            else:
                print(f"{(i,j)} leads to an undefined regualtion state: {matrix.X[i,j]}")
    return(S)


def define_train_and_test_set(S, ratio=0.05):
    D={}
    D['Train']={}
    D['Test']={}

    l=0
    for key in S.keys():
        np.random.seed(l)
        l+=1

        v=S[key]
        np.random.shuffle(v)

        D['Train'][key]=v[:round(len(v)*(1-ratio))]
        D['Test'][key]=v[round(len(v)*(1-ratio)):]
    return(D)



def variant_to_set(Ds, variant):
    
    neg_set=Ds['Not regulated']

    if variant=='up':
        pos_set=Ds['Up-regulated']
        neg_set=[*neg_set, *Ds['Down-regulated']]
    elif variant=='down':
        pos_set=Ds['Down-regulated']
        neg_set=[*neg_set, *Ds['Up-regulated']]
    elif variant=='up+down':
        pos_set=[*Ds['Down-regulated'], *Ds['Up-regulated']]
    else:
        print('Variant not defined (must be in "up", "down", "up+down")')
    return(neg_set, pos_set)



def evaluate_test_set(D, C, variant, plot_hists=True, nn=False):
    not_reg=[]
    reg=[]

    for i in D['Test']['Not regulated']:
        not_reg.append(C[i])

    if variant=='up':
        for i in D['Test']['Up-regulated']:
            reg.append(C[i])
        for i in D['Test']['Down-regulated']:
            not_reg.append(C[i])

    elif variant=='down':
        for i in D['Test']['Down-regulated']:
            reg.append(C[i])
        for i in D['Test']['Up-regulated']:
            not_reg.append(C[i])

    elif variant=='up+down':
        for i in D['Test']['Down-regulated']:
            reg.append(C[i])
        for i in D['Test']['Up-regulated']:
            reg.append(C[i])
    
    df_not_reg=pd.DataFrame({'Score':not_reg, 'Set':'Not regulated'})
    df_reg=pd.DataFrame({'Score':reg, 'Set':f'Regulated ({variant})'})
    df=pd.concat([df_not_reg, df_reg])
    df.index=np.arange(len(df))
    
    if plot_hists:
        
        l=[]
        for i in D['Test']['Unknown']:
            l.append(C[i])
        df_unknowns=pd.DataFrame({'Distribution of scores for the different sets':l, 'Set':'Unknown'})


        sns.set(rc={'figure.figsize':(12,5)})
        sns.set_style('darkgrid')

        plt.subplot(2, 1, 1)

        sns.set(rc={'figure.figsize':(12,5)})
        
        
        if nn:
            bins=np.linspace(0.2,1.2,101)
        else:
            bins=np.linspace(-1.7,1.2,101)
        
        sns.histplot(data=df, x='Score', hue='Set', log_scale=(False, False),bins=bins)
        plt.ylim(0, 300)

        plt.subplot(2, 1, 2)

        sns.histplot(data=df_unknowns, x='Distribution of scores for the different sets', hue='Set', bins=bins)

    return(df)



def errors_given_cutoff(df, cut):
    df_not_reg=df[df['Set']=='Not regulated']

    variant=df[df['Set']!='Not regulated']['Set'].iloc[0].split('(')[1][:-1]

    df_reg=df[df['Set']==f'Regulated ({variant})']


    if variant!='down':

        false_positives=np.array(df_not_reg['Score']>cut).sum()
        true_negatives=np.array(df_not_reg['Score']<=cut).sum()

        false_negatives=np.array(df_reg['Score']<cut).sum()
        true_positives=np.array(df_reg['Score']>=cut).sum()

    else:
        false_positives=np.array(df_not_reg['Score']<cut).sum()
        true_negatives=np.array(df_not_reg['Score']>=cut).sum()

        false_negatives=np.array(df_reg['Score']>cut).sum()
        true_positives=np.array(df_reg['Score']<=cut).sum()

    precission=true_positives/(true_positives+false_positives)
    recall=true_positives/(true_positives+false_negatives)

    return(precission, recall)

def precission_recall_curve(df, bins=201, plot=True, zoom_in=False):
    c=[]
    p=[]
    r=[]

    for cut in np.linspace(-1.5,1.5,bins):

        precission, recall = errors_given_cutoff(df, cut)

        if not np.isnan(precission) and not np.isnan(recall):
            c.append(cut)
            p.append(precission)
            r.append(recall)

    df_precission_recall=pd.DataFrame({'Cutoff':c, 'Precission':p, 'Recall':r, 'Method':'ALS'})
    
    if plot:
    
        sns.set_style("darkgrid")
        sns.set(rc={'figure.figsize':(9,6)})

        fig = plt.figure()

        sns.lineplot(data=df_precission_recall, x='Recall', y='Precission', hue='Method')

        if zoom_in:
            plt.ylim(0.7, 1.0)
            plt.xlim(0.7, 1.0)
        else:
            plt.ylim(-0.05, 1.05)
            plt.xlim(-0.05, 1.05)
        
    return(df_precission_recall)



def plot_precision_recall_comparison(matrix):
    df0=matrix.uns['ALS']
    df0['Method']='ALS'
    df1=matrix.uns['NN_Additional_features']
    df1['Method']='NN_Additional_features'
    dfc=pd.concat([df0, df1])
    dfc.index=np.arange(len(dfc))
    
    sns.lineplot(data=dfc, x='Recall', y='Precission', hue='Method')
    plt.show()
    return(dfc)


def run_weighted_ALS(matrix, variant='up+down', test_set_ratio=0.05, dims=20, regularization=10, 
                     iterations=50, weight_empty=0.01, weight_unknown=0.01):
    S=partition_matrix(matrix)
    D=define_train_and_test_set(S, ratio=test_set_ratio)
    X=construct_train_matrix(matrix, D, variant=variant)
    C,X0,Y0=als.run_ALS_with_weights(matrix, D, X, dims, regularization, iterations, 
                                 weight_empty, weight_unknown)
    df=evaluate_test_set(D,C,variant)
    df_pr=precission_recall_curve(df, bins=10**3, plot=False, zoom_in=False)
    plt.pause(0.1)
    matrix.layers['ALS']=C
    matrix.uns['ALS']=df_pr
    matrix.varm['Drug_embedding']=Y0.T
    matrix.obsm['Peptide_embedding']=X0.T
    return(matrix)


def run_NN_RS(matrix, variant='up+down', test_set_ratio=0.05, epochs=3*10**2, dim0=20, dim1=5, dim_dl=3, dim_features=2, dim_dl_and_features=3):
    from tensorflow.keras import models, layers, utils  #(2.6.0)
    S=partition_matrix(matrix)
    D=define_train_and_test_set(S, ratio=test_set_ratio)   
    in0, in1, out=restructure_train_set(D, variant)
    EC50, SN, FC=nn.get_additional_information(matrix, in0, in1)
    model=nn.mf_and_dl_and_additional_features(matrix, dim0, dim1, dim_dl, dim_features, dim_dl_and_features)
    utils.plot_model(model, to_file='model.png', show_shapes=True, show_layer_names=True)
    training = model.fit(x=[in0, in1, EC50, SN, FC], y=out, epochs=3*10**2, batch_size=10**6, shuffle=True, verbose=0, validation_split=0.1)   
    C=nn.model_to_output_test_set(matrix, D, model, test_only=False) 
    df=evaluate_test_set(D,C,variant, nn=True)
    df_pr=precission_recall_curve(df, bins=10**3, plot=False, zoom_in=False)
    plt.pause(0.1)
    matrix.layers['NN_Additional_features']=C
    matrix.uns['NN_Additional_features']=df_pr
    return(matrix, model)




def construct_train_matrix(matrix, D, variant, neg=-1, pos=1):
    n0,n1=matrix.shape
    X=np.zeros((n0,n1))
    
    Ds=D['Train']
    neg_set, pos_set=variant_to_set(Ds, variant)       

    M=matrix.X
    for i in neg_set:
        X[i]=neg
    for i in pos_set:
        X[i]=pos
    return(X)

def construct_train_matrix_correlation(matrix, D, variant, key, initialization):
    n0,n1=matrix.shape
    
    Ds=D['Train']
    neg_set, pos_set=variant_to_set(Ds, variant)     
    
    X=np.full((n0,n1), initialization, dtype=float)
    
    if key in matrix.layers.keys():
        if key!='pEC50' and key!='log(Fold_change_predicted)':
            for i in [*neg_set, *pos_set]:
                X[i]=np.log(float(matrix.layers[key][i]))
        else:
            for i in [*neg_set, *pos_set]:
                X[i]=float(matrix.layers[key][i])
    else:
        print(f'Error: key {key} not in matrix.layers.keys()')

    return(X)



def restructure_train_set(D, variant, neg=0.5, pos=1):
    neg_set, pos_set=variant_to_set(D['Train'], variant)

    in0=[]
    in1=[]
    out=[]


    Dt=D['Train']

    for i in neg_set:
        in0.append(i[0])
        in1.append(i[1])
        out.append(neg)

    for i in pos_set:
        in0.append(i[0])
        in1.append(i[1])
        out.append(pos)

    in0=np.array(in0)
    in1=np.array(in1)
    out=np.array(out)
    p = np.random.permutation(len(in0))
    
    in0=in0[p]
    in1=in1[p]
    out=out[p]
    
    return(in0, in1, out)


import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import scanpy as sc
import scvelo as scv
import statsmodels.stats.multitest
import scipy
import matplotlib.pyplot as plt


def calculate_leiden_peptides(matrix, resolution=2, plot=True):
    Emb=matrix.uns['Peptide_emb_2d']
    sc.pp.neighbors(Emb)
    sc.tl.leiden(Emb, resolution=resolution)
    
    if plot:
        scv.pl.scatter(Emb, color='leiden', size=50, legend_loc='right', palette=sns.color_palette('tab20'))   
    matrix.obs['leiden']=Emb.obs['leiden']
    return(matrix)


def calculate_leiden_drugs(matrix, resolution=2, plot=True):
    Emb=matrix.uns['Drug_emb_2d']
    sc.pp.neighbors(Emb)
    sc.tl.leiden(Emb, resolution=resolution)
    
    if plot:
        scv.pl.scatter(Emb, color='leiden', legend_loc='right', palette=sns.color_palette('tab20'))   
    matrix.var['leiden']=Emb.obs['leiden']
    return(matrix)

def most_ALS_extreme_peptides(matrix, k=500, plot_hist=False):
    Emb=matrix.obsm['Peptide_embedding']

    D={}
    for i in range(Emb.shape[1]):
        emb_i=Emb[:,i]
        D[i]=np.argpartition(emb_i,-k)[-k:]


        if plot_hist:
            mn=min(emb_i[D[i]])        
            sns.histplot(x=emb_i, bins=101)

            plt.plot([mn,mn], [0,1000], color='brown', marker='o', linestyle='dashed', linewidth=2, markersize=5)
            plt.pause(0.01)
    return(D)

def most_ALS_extreme_drugs(matrix, k=20, plot_hist=False):
    Emb=matrix.varm['Drug_embedding']

    D={}
    for i in range(Emb.shape[1]):
        emb_i=Emb[:,i]
        D[i]=np.argpartition(emb_i,-k)[-k:]


        if plot_hist:
            mn=min(emb_i[D[i]])        
            sns.histplot(x=emb_i, bins=101)

            plt.plot([mn,mn], [0,1000], color='brown', marker='o', linestyle='dashed', linewidth=2, markersize=5)
            plt.pause(0.01)
    return(D)


def leiden_peptides(matrix):
    
    matrix.obs['leiden']=matrix.obs['leiden'].astype('category')
    df=matrix.obs['leiden']
    leidens=[a for a in df.cat.categories]

    D={}
    for i in range(len(leidens)):
        D[i]=np.where(df==leidens[i])[0]

    return(D)


def leiden_drugs(matrix):
    
    matrix.var['leiden']=matrix.var['leiden'].astype('category')
    df=matrix.var['leiden']
    leidens=[a for a in df.cat.categories]

    D={}
    for i in range(len(leidens)):
        D[i]=np.where(df==leidens[i])[0]

    return(D)

def load_kegg_matrix(matrix, signaling_only=True):
    df=matrix.uns['KEGG_matrix']
    
    if signaling_only:
        # Keep only pathways that contain "signal" in it's name

        keep=[]

        for i in range(len(df.T)):
            clas=df.columns[i]
            if 'signal' in clas:
                keep.append(clas)
        df=df[keep].copy()
        
    return(df)

def load_drug_class_matrix(matrix):
    df=matrix.uns['Drug_class_matrix']
    dfs=df.T[df.sum()>1].T
        
    return(dfs)



def binomial_test(D, df, mulititest=True):
    Base_prob={}
    for i in range(len(df.T)):
        clas=df.columns[i]
        Base_prob[clas]=df.sum()[i]/len(df)


    P=np.zeros((len(D), df.shape[1]))
    for i in range(len(D)):
        occurrences=df.iloc[D[i]].sum()
        for j in range(len(df.columns)):
            clas=[a for a in df.columns][j]             
            found=occurrences[clas]
            base_prob=Base_prob[clas]

            P[i,j]=scipy.stats.binomtest(found, len(D[i]), p=base_prob, alternative='greater').pvalue


    p=P.flatten()
    q=statsmodels.stats.multitest.multipletests(p, alpha=0.05, method='fdr_bh')[1]
    Q=q.reshape(P.shape)

    if mulititest:
        Pdf=pd.DataFrame(Q)
    else:
        Pdf=pd.DataFrame(P)
        
    Pdf.index=[a for a in np.arange(len(D.keys()))]
    Pdf.columns=df.columns
    return(Pdf)


def calc_clustermap(Pdf):
    cluster_map=sns.clustermap(Pdf, figsize=(0.0001, 0.0001), cmap='magma_r', vmin=0, vmax=0.1)
    plt.axis('off')
    cluster_map._figure=None
    col_reorder=cluster_map.dendrogram_col.reordered_ind
    row_reorder=cluster_map.dendrogram_row.reordered_ind
    
    return(col_reorder, row_reorder)



def print_sub_heatmap(Pdf, col_reorder, row_reorder, i0, i1=0):
    if i1==0:
        i1=i0+50
    cols=[[a for a in Pdf.columns][i] for i in col_reorder]
    rows=[[a for a in Pdf.index][i] for i in row_reorder]
    
    Pdfn=Pdf.copy()
    
    Pdfn=Pdfn[cols]
    Pdfn=Pdfn.T
    Pdfn=Pdfn[rows]
    Pdfn=Pdfn.T
    Pdfn.index=[str(i) for i in Pdfn.index]
    
    plt.figure(figsize = (20,10), dpi=100)
    sns.heatmap(Pdfn.T[i0:i1].T, cmap='magma_r', vmin=0, vmax=0.1)
    
    
def cut_of_p_values(Pdf, cutoff=0.1):
    x,y=np.where(Pdf>cutoff)
    for i in range(len(x)):
        Pdf.values[x[i],y[i]]=1
    Pdf=Pdf.T[Pdf.sum()<Pdf.shape[0]-0.92].T
    return(Pdf)




def calc_peptide_heatmap(matrix, clustering='Leiden', signaling_only=True, cutoff=0.1):
    if clustering=='ALS':
        D=most_ALS_extreme_peptides(matrix)
    elif clustering=='Leiden':
        D=leiden_peptides(matrix)
    else:
        print(f'Error: Unknown clustering (not in (ALS, Leiden))')
    df=load_kegg_matrix(matrix, signaling_only=signaling_only)
    Pdf=binomial_test(D,df)
    Pdf=cut_of_p_values(Pdf, cutoff=cutoff)
    col_reorder, row_reorder=calc_clustermap(Pdf)
    print_sub_heatmap(Pdf, col_reorder, row_reorder, i0=0, i1=200)
    
    return(Pdf)


def calc_drug_heatmap(matrix, clustering='Leiden', cutoff=0.1):
    if clustering=='ALS':
        D=most_ALS_extreme_drugs(matrix)
    elif clustering=='Leiden':
        D=leiden_drugs(matrix)
    else:
        print(f'Error: Unknown clustering (not in (ALS, Leiden))')
    df=load_drug_class_matrix(matrix)
    Pdf=binomial_test(D,df, mulititest=False)
    Pdf=cut_of_p_values(Pdf, cutoff=cutoff)
    col_reorder, row_reorder=calc_clustermap(Pdf)
    print_sub_heatmap(Pdf, col_reorder, row_reorder, i0=0, i1=200)
    
    return(Pdf)

    


import pandas as pd
import numpy as np
import seaborn as sns
import anndata

def list_to_str(lst):
    string=''
    for i in lst:
        string+=str(i)+', '
    string=string[:-2]
    return(string)

def ratio_overview(matrix, pr=False):
    n=0
    z=0
    un=0
    up=0
    down=0
    m=len(matrix)*len(matrix.var)

    for i in range(len(matrix)):
        for j in range(len(matrix.var)):
            if matrix.X[i,j]=='Empty':
                n+=1
            elif matrix.X[i,j]=='Unknown':
                un+=1
            elif matrix.X[i,j]=='Not regulated':
                z+=1
            elif matrix.X[i,j]=='Up-regulated':
                up+=1
            elif matrix.X[i,j]=='Down-regulated':
                down+=1
            else:
                print(f'Classification class not found: {matrix.X[i,j]}')
                
                
    if pr:
        print(f'Up-regulated: {round(up/m*10000)/100}%')
        print(f'Down-regulated: {round(down/m*10000)/100}%')
        print(f'Not regulated: {round(z/m*10000)/100}%')
        print(f'Unknown regulation: {round(un/m*10000)/100}%')
        if n>=1:    
            print(f'No curve: {round(n/m*10000)/100}%')
            
    df=pd.DataFrame({'Up-reg.':[up/m], 'Down-reg.':down/m, 'Not reg.':z/m, 'Unknown':un/m, 'No curve':n/m, 'Total amount of regulated p-sites':up+down})
        
    return(df)


def filter_drugs(matrix, min_reg_phos=10):
    # min_reg_phos = min_regulated_psites_per_drug 
    sel=[]
    for i in range(len(matrix.var)):
        k=0
        for j in range(len(matrix)):
            entry=matrix.X[j,i]
            if entry=='Unknown' or entry=='Not regulated' or entry=='Empty':
                continue
            else:
                k+=1
        if k>min_reg_phos-1:
            sel.append(i)
    return(matrix[:,sel])

def filter_peptides(matrix, min_reg_pept=3):
    # min_reg_pept = min_drugs_per_regulated_psite
    sel=[]
    for i in range(len(matrix)):
        k=0
        for j in range(len(matrix.var)):
            entry=matrix.X[i,j]
            if entry=='Unknown' or entry=='Not regulated' or entry=='Empty':
                continue
            else:
                k+=1
        if k>min_reg_pept-1:
            sel.append(i)
    return(matrix[sel,:])


def load_matrix(path, cell_line='A204'):
    Matrix=pd.read_csv(f'{path}', sep='\t', on_bad_lines='skip')

    obs=Matrix[['Gene Names', 'Proteins', 'Modified sequence']]
    obs.index=[str(a) for a in obs.index]
    Matrix=Matrix.drop(['Gene Names', 'Proteins', 'Modified sequence', 'Sequence'], axis=1)
    drugs=[a for a in Matrix.columns]
    X=np.array(Matrix.values, dtype=np.float32)
    n0,n1=X.shape

    Xn=np.full(X.shape, '-'*20, dtype='<U20')
    for i in range(n0):
        for j in range(n1):
            if np.isnan(X[i,j]):
                Xn[i,j]='Unknown'
            elif X[i,j]==0:
                Xn[i,j]='Not regulated'
            elif X[i,j]<0:
                Xn[i,j]='Down-regulated'
            elif X[i,j]>0:
                Xn[i,j]='Up-regulated'

    matrix=anndata.AnnData(X=Xn, obs=obs, dtype=Xn.dtype)
    matrix.var['Drug']=drugs
    return(matrix.copy())


def add_drug_names_and_classes(matrix, Path):
    # Translate TOPAS-ID to drug name and add drug classes
    Drug_name_dict=np.load(f'{Path}/TOPAS_Drugs_A204.npy', allow_pickle=True)
    Drug_name_dict=dict(enumerate(Drug_name_dict.flatten()))[0]

    Drug_class_dict=np.load(f'{Path}/Drug_classes.npy', allow_pickle=True)
    Drug_class_dict=dict(enumerate(Drug_class_dict.flatten()))[0]

    Drug_class_combined_dict=np.load(f'{Path}/Drug_classes_combined.npy', allow_pickle=True)
    Drug_class_combined_dict=dict(enumerate(Drug_class_combined_dict.flatten()))[0]
    
    
    matrix.var['Drug Names']='-'*20
    matrix.var['Drug Class']='-'*20
    matrix.var['Drug Class combined']='-'*20
    
    for i in range(len(matrix.var)):
        drug=matrix.var['Drug'][i]
        matrix.var['Drug Names'][i]=Drug_name_dict[drug]
        matrix.var['Drug Class'][i]=Drug_class_dict[Drug_name_dict[drug]]
        matrix.var['Drug Class combined'][i]=Drug_class_combined_dict[Drug_name_dict[drug]]

    return(matrix.copy())

def add_Kegg_annontations(matrix, Path):
    # Matthew provided me with Kegg pathway annotations, they are added here to the respective peptides
    
    # Load the Dictionaries:

    KEGG_gene=np.load(f'{Path}/KEGG_genes.npy', allow_pickle=True)
    KEGG_gene=dict(enumerate(KEGG_gene.flatten()))[0]
    
    KEGG_protein=np.load(f'{Path}/KEGG_proteins.npy', allow_pickle=True)
    KEGG_protein=dict(enumerate(KEGG_protein.flatten()))[0]
    
    kegg_gene=[['-']] * len(matrix)
    kegg_protein=[['-']] * len(matrix)
    kegg_combinded=[['-']] * len(matrix)

    for i in range(len(matrix)):
        gene=matrix.obs['Gene Names'][i]
        protein=matrix.obs['Proteins'][i]
        
        if gene in KEGG_gene.keys():
            kegg_gene[i]=KEGG_gene[gene]
        else:
            kegg_gene[i]=['-']
            
        if protein in KEGG_protein.keys():
            kegg_protein[i]=KEGG_protein[protein]
        else:
            kegg_protein[i]=['-']
            
        KEGG_comb=[*kegg_gene[i], *kegg_protein[i]]
        KEGG_comb=[a for a in set(KEGG_comb)]
        
        if KEGG_comb=='-':
            kegg_combinded[i]=['-']
        else:
            KEGG_comb.remove('-')
            kegg_combinded[i]=KEGG_comb
            
    matrix.obs['KEGG_gene']=[list_to_str(a) for a in kegg_gene]
    matrix.obs['KEGG_protein']=[list_to_str(a) for a in kegg_protein]
    matrix.obs['KEGG_combined']=[list_to_str(a) for a in kegg_combinded]
        
    return(matrix.copy())


def add_Phosphosite_annotations(matrix, psite_annotation_path):
    
    import sys
    sys.path.append(psite_annotation_path)
    import psite_annotation as pa
    
    df=matrix.obs.copy()    
    
    # Adds effect of phosphorylation
    pa.addPeptideAndPsitePositions(df, pa.pspFastaFile, pspInput = True)
    pa.addPSPAnnotations(df, pa.pspAnnotationFile)
    pa.addPSPRegulatoryAnnotations(df, pa.pspRegulatoryFile)
    pa.addPSPKinaseSubstrateAnnotations(df, pa.pspKinaseSubstrateFile)
    #pa.addDomains(df, pa.domainMappingFile)
    #pa.addMotifs(df, pa.motifsFile)
    pa.addInVitroKinases(df, pa.inVitroKinaseSubstrateMappingFile)
    

    
    matrix.obs=df.copy()
    return(matrix.copy())

def add_Yasushi_annotations(matrix, Path):
    
    Yas_dict=np.load(f'{Path}/Yasushi_annotations.npy', allow_pickle=True)
    Yas_dict=dict(enumerate(Yas_dict.flatten()))[0]

    yasushi=[['-']] * len(matrix)
    
    for i in range(len(matrix)):
        gene=matrix.obs['Gene Names'][i]
        if gene in Yas_dict.keys():
            yasushi[i]=list_to_str(Yas_dict[gene])
        else:
            yasushi[i]='-'
    matrix.obs['Yasushi']=yasushi
    return(matrix)

def upstream(df, key, Path):
    Tr=np.load(f'{Path}/Gene_name_translator.npy', allow_pickle=True)
    Tr=dict(enumerate(Tr.flatten()))[0]
    
    if key=='vivo':
        key2='PSP'
    if key=='vitro':
        key2='In Vitro'
    
    df[f'Upstream_{key}']=[a for a in df[f'{key2} Kinases']]
    upstream=['']*len(df)
    
    for i in range(len(df)):
        s=df[f'{key2} Kinases'][i]
        ls=s.split(';')

        if len(s)>1:
            for j in range(len(ls)):
                st=ls[j]
                if '/' in st:
                    st=st.split('/')[0]
                if '[' in st:
                    st=st.split('[')[0]
                if st in Tr.keys():
                    st=Tr[st]
                upstream[i]+=st+', '

    up2=[a[:-2] for a in upstream]

    up2_new=np.full(len(up2), '', dtype='<100U')
    for i in range(len(up2)):
        for a in set(up2[i].split(', ')):
            if a!='':
                up2_new[i]+=', '+a
        if up2_new[i]!='':
            up2_new[i]=up2_new[i][2:]
                
                
    df[f'Upstream_{key}']=up2_new
      
    return(df)

def add_up_and_downstream_annotations(matrix, Path):
    df=matrix.obs.copy()
    
    
    # Adds downstream annotation
    
    down_act=[''] * len(matrix)
    down_inh=[''] * len(matrix)
    down_unk=[''] * len(matrix)

    for i in range(len(matrix)):

        s=df['PSP_ON_PROT_INTERACT'][i]

        if s!='-':
            ls=df['PSP_ON_PROT_INTERACT'][i].split('; ')

            for j in range(len(ls)):

                if ls[j][-9:]=='(INDUCES)':
                    if down_act[i]!='':
                        down_act[i]+=', '
                    down_act[i]+=ls[j][:-9]
                elif ls[j][-10:]=='(DISRUPTS)':
                    if down_inh[i]!='':
                        down_inh[i]+=', '
                    down_inh[i]+=ls[j][:-10]
                elif ls[j][-14:]=='(NOT_REPORTED)':
                    if down_unk[i]!='':
                        down_unk[i]+=', '
                    down_unk[i]+=ls[j][:-14]
                    
    # Adds upstream annotation
                    
    df=upstream(df, 'vivo', Path)
    df=upstream(df, 'vitro', Path)
    
    df['Downstream_activated']=down_act
    df['Downstream_inhibited']=down_inh
    df['Downstream_unknown']=down_unk
    
    matrix.obs=df.copy()
    
    return(matrix.copy())

def apply_basic_filtering(matrix):
    matrix=filter_drugs(matrix).copy()
    matrix=filter_peptides(matrix).copy()
    return(matrix)


def create_drug_class_matrix(Path, matrix, cutoff=1000):
    kinobead_data=pd.read_csv(f'{Path}/drug_target_matrix.tsv', sep='\t')
    drugs=kinobead_data['Drug']
    kinobead_data.index=drugs
    kinobead_data=kinobead_data.drop(['Drug','Unnamed: 1'], axis=1)

    df=kinobead_data.copy()

    n0,n1=df.shape

    bools=df.values>cutoff
    bools=np.array(bools, dtype=int)

    dfn=pd.DataFrame(bools, columns = df.columns)
    dfn.index=kinobead_data.index

    dfn=dfn.T[matrix.var['Drug Names']].T.copy()
    return(dfn, kinobead_data.T[dfn.index].T)

def create_KEGG_matrix(matrix):
    # Get all classes of KEGG_combined
    Dx={}
    for i in range(len(matrix)):
        KEGG_comb=matrix.obs['KEGG_combined'][i]

        if KEGG_comb!='':
            elements=matrix.obs['KEGG_combined'][i].split(',')        
            for e in elements:
                if e!='':
                    if e[0]==' ':
                        e=e[1:]
                if not e in Dx.keys():
                    Dx[e]=len(Dx)

    KEGG_matrix=np.zeros((len(matrix),len(Dx)), dtype=int)

    for i in range(len(matrix)):
        KEGG_comb=matrix.obs['KEGG_combined'][i]

        if KEGG_comb!='':
            elements=matrix.obs['KEGG_combined'][i].split(',')        
            for e in elements:
                if e!='':
                    if e[0]==' ':
                        e=e[1:]
                if e in Dx.keys():
                    KEGG_matrix[i,Dx[e]]=1
                else:
                    print(e)  
    df = pd.DataFrame(KEGG_matrix, columns = [a for a in Dx.keys()])
    df.index=matrix.obs['Modified sequence']
    return(df)


def cleanup(matrix):
    df=matrix.obs.copy()
    for i in ['Start positions', 'End positions', 'Site sequence context', 'Site positions', 'PSP_LT_LIT', 
              'PSP_MS_LIT', 'PSP_MS_CST', 'PSP_ON_FUNCTION', 'PSP_ON_PROCESS', 'PSP_ON_PROT_INTERACT', 
              'PSP_ON_OTHER_INTERACT', 'PSP_NOTES', 'PSP Kinases', 'In Vitro Kinases']:
        del df[i]
        
    for i in ['KEGG_gene', 'KEGG_protein', 'KEGG_combined', 'Yasushi', 'Downstream_activated', 
              'Downstream_inhibited', 'Downstream_unknown', 'Upstream_vivo', 'Upstream_vitro']:
        arr=[''*200]*len(matrix)
        for j in range(len(matrix)):
            if df[i][j]=='':
                arr[j]='-'
            else:
                arr[j]=df[i][j]
        df[i]=arr
    matrix.obs=df.copy()
    
    matrix.var.index=np.arange(len(matrix.var))
    matrix.obs.index=np.arange(len(matrix))
    
    matrix.obs.index=[str(a) for a in matrix.obs.index]
    matrix.var.index=[str(a) for a in matrix.var.index]
    
    return(matrix.copy())

def colors_based_on_phylogeny(matrix, Path):
    import matplotlib.colors as colors

    Drug_colors=np.load(f'{Path}/Drug_classes_colors.npy', allow_pickle=True)
    Drug_colors=dict(enumerate(Drug_colors.flatten()))[0]

    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_2d_umap'], obs=matrix.var)
    Emb_matrix.obsm['X_umap']=matrix.varm['Drug_2d_umap']


    col=[]
    for i in Emb_matrix.obs['Drug Class combined'].cat.categories:
        col.append(Drug_colors[i])

    col2=[]
    for i in Emb_matrix.obs['Drug Class combined']:
        if not colors.rgb2hex(Drug_colors[i]) in col2:
            col2.append(colors.rgb2hex(Drug_colors[i]))
    return(col, col2)



def plot_summary(before_filtering, after_filtering):
    import matplotlib.pyplot as plt
    import numpy as np
    
    before_filtering['When']='Before filtering'
    after_filtering['When']='After filtering'
    comb=pd.concat([before_filtering, after_filtering])
    comb.index=[0,1]

    classes=comb.columns[:5]

    df0=pd.DataFrame({'Percentage': before_filtering.values[0][:5]*100, 'Classes': classes, 'When':'Before filtering'})
    df1=pd.DataFrame({'Percentage': after_filtering.values[0][:5]*100, 'Classes': classes, 'When':'After filtering'})

    DF=pd.concat([df0, df1])
    DF.index=np.arange(len(DF))


    sns.set_style('darkgrid')

    fig, axs = plt.subplots(1,2,figsize=(10,4), gridspec_kw={'width_ratios': [1, 2]})

    #### Subplot 1
    plt.subplot(1, 2, 1)

    sns.barplot(data=comb, x='When', y='Total amount of regulated p-sites')

    #### Subplot 2
    plt.subplot(1, 2, 2)
    sns.barplot(data=DF, x='Classes', y='Percentage', hue='When')

    plt.show()
    
def add_kinobead_and_kegg_matrices(matrix, Path):
    matrix.uns['Drug_class_matrix'], matrix.uns['Kinobead_matrix']=create_drug_class_matrix(Path, matrix, cutoff=1000)
    matrix.uns['KEGG_matrix']=create_KEGG_matrix(matrix)
    return(matrix)




def minimize_site_positions(matrix):
    df=matrix.obs.copy()
    df['Index']=np.arange(len(df))
    df['Gene Names']=df['Gene Names'].astype('category')
    genes=df['Gene Names'].cat.categories

    frames=[]
    for gene in genes:
        dfg=df[df['Gene Names']==gene].copy()
        dfg=reduce_site_positions(dfg)
        frames.append(dfg)

    dfc=pd.concat(frames)
    dfc=dfc.sort_values('Index').copy()

    matrix.obs['Reference_protein']=dfc['Reference_protein']
    matrix.obs['Ph_site']=dfc['Ph_site']
    
    return(matrix)


def identify_most_common_protein(dfg):
    prots=[]
    for i in range(len(dfg)):
        sites=dfg['Site positions'][i].split(',')
        for site in sites:
            prots.append(site.split('_')[0])
    prots=[a for a in set(prots)]

    if prots==['']:
        most_common='-'

    else:
        D={}
        for prot in prots:
            if prot=='':
                continue
            else:
                z=np.zeros(len(dfg))
                for i in range(len(dfg)):
                    if prot in dfg['Site positions'][i]:
                        z[i]=1
                D[prot]=[z.sum()]
        dtf=pd.DataFrame(D).T

        most_common=dtf.sort_values(0).index[-1]

    return(most_common)

def reduce_site_positions(dfg):
    com=identify_most_common_protein(dfg)

    site_pos=[]
    for i in range(len(dfg)):
        entry='-'
        sites=dfg['Site positions'][i].split(';')
        for site in sites:
            if com in site:
                if entry=='-':
                    entry=site.split('_')[1]
                else:
                    entry+='_'+site.split('_')[1]
        site_pos.append(entry)
    dfg['Reference_protein']=com
    dfg['Ph_site']=site_pos
    return(dfg)

    
    
    
def preprocess(matrix, Path, psite_annotation_path, min_reg_phos=10, min_reg_pept=3):
    matrix=add_drug_names_and_classes(matrix, Path)
    matrix=add_Kegg_annontations(matrix, Path)
    matrix=add_Phosphosite_annotations(matrix, psite_annotation_path).copy()
    matrix=minimize_site_positions(matrix)
    matrix=add_Yasushi_annotations(matrix, Path)
    matrix=add_up_and_downstream_annotations(matrix, Path)
    before_filtering=ratio_overview(matrix, pr=False)
    matrix=filter_drugs(matrix, min_reg_phos).copy()
    matrix=filter_peptides(matrix, min_reg_pept).copy()
    after_filtering=ratio_overview(matrix, pr=False)
    plot_summary(before_filtering, after_filtering)
    matrix=cleanup(matrix.copy())
    matrix=add_kinobead_and_kegg_matrices(matrix, Path)
    return(matrix)
    


import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import matplotlib.pyplot as plt
import scipy

import os
import sys
sys.path.append(os.getcwd().rsplit('/', 1)[0]+'/Functions')

import Extract_from_raw_data as extr


doses = [0.0, 0.03, 0.3, 1.0, 3.0, 10.0, 30.0, 100.0, 300.0, 1000.0, 10000.0]


def cutoff_given_precission(matrix, key, min_precission):
    df=matrix.uns[key]
    precission=np.array([a for a in df['Precission']])

    pr2=precission-min_precission
    pr2=[1 if a<0 else a for a in pr2]
    ind=np.argmin(pr2)

    cutoff, prec, recall=[a for a in df.iloc[ind]][:3]
    return(cutoff, prec, recall)

def recommend_peptide_drug_pairs(matrix, key, min_precision):
    cutoff, prec, recall=cutoff_given_precission(matrix, key, min_precision)
    
    n0,n1=matrix.shape
    
    recommended=[]
    for i in range(n0):
        for j in range(n1):
            if matrix.layers[key][i,j]>cutoff:
                recommended.append((i,j))
    return(recommended)


def select_only_identified(matrix, recommended):
    identified=[]

    for i in recommended:
        if matrix.X[i]=='Down-regulated' or matrix.X[i]=='Up-regulated' or matrix.X[i]=='Not regulated':

            dr=matrix.var.iloc[i[1]]['Drug']
            pept=matrix.obs.iloc[i[0]]['Modified sequence']
            df=Data[dr]

            if pept in df['Modified sequence']:
                identified.append(i)
    return(identified)


def get_curves(matrix, Data, indices, pick_criterion='Andromeda Score'):
    
    frames=[]
    
    if 'Empty' in matrix.X:
        for i in indices:
            if matrix.X[i]!='Empty':
                
                dr=matrix.var.iloc[i[1]]['Drug']
                pept=matrix.obs.iloc[i[0]]['Modified sequence']
                df=Data[dr]

                dfs=df[df['Modified sequence']==pept]
                dfs=extr.pick_curve(dfs, pick_criterion=pick_criterion).copy()
                frames.append(dfs)
        
    else:
        for i in indices:
            dr=matrix.var.iloc[i[1]]['Drug']
            pept=matrix.obs.iloc[i[0]]['Modified sequence']
            df=Data[dr]

            if pept in [a for a in df['Modified sequence']]:
                dfs=df[df['Modified sequence']==pept]
                dfs=extr.pick_curve(dfs, pick_criterion=pick_criterion).copy()
                frames.append(dfs)

    Df=pd.concat(frames)
    Df.index=np.arange(len(Df))
    Df['log2(new_FC)']=np.log2(Df['Fold_change_predicted'])
    return(Df)



def regulation_for_recommended_curves(df, alpha=0.03, fc_lim=0.35, new_not_regualted=True):

    Fold_change_key='Fold_change_predicted'
    Q_value_key='Curve Log P_Value'

    ############## This is a reduced version of Flo's "define_regulated_curves"    ##################################

    fold_change = np.log2(df[Fold_change_key].copy())
    s0 = extr.get_s0(fc_lim, alpha)
    p_value_cutoff = extr.map_fc_to_p_samcutoff(fold_change, alpha=alpha, s0=s0, deg_f=10-2)
    p_mask = df[Q_value_key] >= p_value_cutoff
    down_mask = fold_change < 0
    up_mask = fold_change > 0

    # Not regulated curves require: not too much noise, not too much deviation to control
    # TODO: Do we need a p value cut off ? currently up and down regulation will overwrite not,
    # also p value might be strong but no effect
    rmse_max = 0.1
    
    if new_not_regualted:
        not_regulated_mask = (df['Curve Simple Model RMSE'] < rmse_max) & \
                             (np.log2(df[Fold_change_key])).between(-fc_lim, fc_lim)
    
    else:
        not_regulated_mask = (df['Curve Simple Model RMSE'] < rmse_max) & \
                             (np.log2(df['Curve Simple Model'])).between(-abs(fc_lim)/2, abs(fc_lim)/2)


    # Add labels
    df['Regulation_after_RS'] = 'Unknown'
    df.loc[not_regulated_mask, 'Regulation_after_RS'] = 'Not regulated'
    df.loc[p_mask & up_mask, 'Regulation_after_RS'] = 'Up-regulated'
    df.loc[p_mask & down_mask, 'Regulation_after_RS'] = 'Down-regulated'
    
    return(df)


def vulcano_plot_recommender_curves(DF, alpha=0.03, fc_lim=0.35, plot_type='scatter', color='New_curve_regulation'):
    sns.set_style('darkgrid')
    sns.set(rc={'figure.figsize':(12,5)})

    deg_f=10-2
    x=np.linspace(fc_lim,7,100)
    y=extr.sam_curve(x, deg_f, fc_lim, alpha)

    x=[*-x,*x]
    y=[*y,*y]
    
    if plot_type=='scatter':
        # coloring with Drug and Confidence seems to only work with scatter plot and not with kde
        
        sns.scatterplot(data=DF, y='Curve Log P_Value', x='log2(new_FC)', s=10, hue='New_curve_regulation')
        
    elif plot_type=='kde':
        sns.kdeplot(data=DF, y='Curve Log P_Value', x='log2(new_FC)', s=10, hue='New_curve_regulation', cmap="Blues", shade=True, bw_adjust=0.2)
    
    else:
        print('Error: plot_type is not in [scatter, kde]')
    
    sns.lineplot(x=x,y=y, color='brown')
    
    
def apply_thresholds(Df, fc_lim=0.55, alpha=0.015, plot=True):
    log_alpha=-np.log10(alpha)

    sns.set_style('darkgrid')
    sns.set(rc={'figure.figsize':(12,5)})

    color='New_curve_regulation'
    Df_rec=Df[abs(Df['log2(new_FC)'])>fc_lim].copy()
    Df_rec=Df_rec[Df_rec['Curve Log P_Value']>log_alpha].copy()
    
    Df['-log10(p-value curve fit)']=Df['Curve Log P_Value']
    Df['log2(Fold change)']=Df['log2(new_FC)']
    Df['Curve regulation']=Df['New_curve_regulation']
    
    
    
    if plot:
        print(f'{round(len(Df_rec)/len(Df)*1000)/10}% of recommended curves ({len(Df_rec)} curves) satisfy the more leniant leniant classification criteria')
        deg_f=10-2
        x=np.linspace(fc_lim,7,1000)
        y=extr.sam_curve(x, deg_f, fc_lim, alpha)

        x=[*-x,*x]
        y=[*y,*y]

        sns.kdeplot(data=Df, y='-log10(p-value curve fit)', x='log2(Fold change)', s=10, hue='Curve regulation', cmap="Blues", shade=True, bw_adjust=1)

        sns.lineplot(x=x,y=y, color='brown')

        sns.lineplot(x=[-fc_lim,-fc_lim], y=[log_alpha, 10], color='brown', estimator=None, linewidth = 2)
        sns.lineplot(x=[fc_lim,fc_lim], y=[log_alpha, 10], color='brown', estimator=None, linewidth = 2)
        sns.lineplot(x=[-3,-fc_lim], y=[log_alpha, log_alpha], color='brown')
        sns.lineplot(x=[fc_lim,3], y=[log_alpha, log_alpha], color='brown')

        plt.ylim(0,8)
        plt.xlim(-3,3)
    
    return(Df_rec)       


def recommend_for_segmented(matrix, key='up', min_precission=0.95):
    c_l,a,b=cutoff_given_precission(matrix, 'dfdo', min_precission)
    c_u,a,b=cutoff_given_precission(matrix, 'dfup', min_precission)

    n0,n1=matrix.shape

    recommend=[]
    for i in range(n0):
        for j in range(n1):
            if matrix.layers['ALS_weighted'][i,j]>c_u:
                recommend.append((i,j))
                
    if key=='down':
        print('Recommending for down-regulation')


        recommend=[]
        for i in range(n0):
            for j in range(n1):
                if matrix.layers['ALS_weighted'][i,j]<c_l:
                    recommend.append((i,j))
    else:
        print('Recommending for up-regulation')
    return(recommend)



def recommend_curves(matrix, Data, rs_used='ALS', min_precision=0.95, unknowns_only=True):
    recommended=recommend_peptide_drug_pairs(matrix, rs_used, min_precision)
    
    if unknowns_only:
        indices=select_only_unknowns(matrix, recommended)
    else:
        indices=recommended
    
    Df=get_curves(matrix, Data, indices, pick_criterion='Andromeda Score')
    
    Dfu=Df[Df['New_curve_regulation']=='Unknown']
    
    Dfr=apply_thresholds(Dfu, plot=True)
    return(Dfr)


def select_only_unknowns(matrix, recommended):
    unknowns=[]
    for i in recommended:
        if matrix.X[i]=='Unknown':
            unknowns.append(i)
    return(unknowns)



import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import matplotlib.pyplot as plt
import scipy


def run_ALS(X, dims=25, regulatization=300, iterations=10**3, alpha=10):
    import implicit
    from scipy import sparse
    n0,n1=X.shape
    
    M = sparse.csr_matrix(X.T)
    model = implicit.als.AlternatingLeastSquares(factors=dims, regularization=regulatization, iterations=iterations, random_state=0)
    alpha = alpha
    data_conf = (M * alpha).astype('double')
    model.fit(data_conf)
    
    
    C=np.zeros((n0,n1))
    for i in range(n1):
        recommended, recommended_scores = model.recommend(i, data_conf[i], filter_already_liked_items=False, N=n0)

        for j in range(n0):
            C[recommended[j],i]=recommended_scores[j]
            
            
    X0=model.item_factors
    Y0=model.user_factors
    X0=gpu_matrix_to_numpy(X0)
    Y0=gpu_matrix_to_numpy(Y0)

    return(C, X0, Y0)

def gpu_matrix_to_numpy(M):
    n0,n1=M.shape

    Dt={}
    for i in range(n0):
        Dt[i]=[]
        st=str(M[i])
        st=st[2:-2]
        a=st.split(' ')

        for e in a:
            Dt[i].append(e.split('\n')[0])
    Dtn={}
    for i in range(n0):
        lst=[]
        for e in Dt[i]:
            if e!='' and e!=' ':
                lst.append(float(e))
        Dtn[i]=np.array(lst)

    Mn=np.zeros(M.shape)
    for i in range(n0):
        if len(Dtn[i])!=n1:
            print(f'Error in gpu_matrix to numpy matrix conversion in row {i}!')
            print(Dtn[i])
        else:
            Mn[i]=Dtn[i]
    return(Mn)

def construct_weight_matrix(matrix, D, key=None):
    n0,n1=matrix.shape
    
    Dt=D['Train']
    train_set=[*Dt['Not regulated'], *Dt['Up-regulated'], *Dt['Down-regulated']]
    
    if key!='pEC50':
        W=np.full((n0,n1), 0.1)
        for i in train_set:
            W[i]=1
    else:
        for i in train_set:
            res=1+(float(matrix.layers['pEC50'][i])/float(matrix.layers['Log EC50 Error'][i]))
            if res<1000:
                W[i]=res
            else:
                W[i]=1000
    return(W)



def run_ALS_with_weights(matrix, D, X, dims=25, regulatization=10, iterations=10**2, weight_empty=0.01, weight_unknown=0.01, key=None):
    n0,n1=X.shape
    
    # Construct weights-matrix
    W=construct_weight_matrix(matrix, D, key=key)
                
    X0,Y0=run_weighted_als(X,W,dims,iterations,regulatization)
        
    C=np.matmul(X0.T,Y0)
    
    return(C,X0,Y0)


def run_weighted_als(X,W,l,iterations,regulatization):
    n0,n1=X.shape   
    
    X1=np.zeros((l,n0))
    Y0=np.random.rand(l,n1)
    Y1=np.zeros((l,n1))
    
    for k in range(iterations):
        for i in range(n0):

            cd=W[i,:]
            pd=X[i,:]

            cY0T=Y0.T * cd[:, np.newaxis]
            M=np.matmul(Y0,cY0T)+regulatization*np.diag(np.ones(l))
            b=np.matmul(Y0,pd*cd)
            X1[:,i]=np.linalg.solve(M,b)
        X0=X1.copy()

        for i in range(n1):

            cp=W[:,i]
            pp=X[:,i]
            cX0T=X0.T * cp[:, np.newaxis]
            M=np.matmul(X0,cX0T)+regulatization*np.diag(np.ones(l))

            b=np.matmul(X0,pp*cp)
            Y1[:,i]=np.linalg.solve(M,b)
        Y0=Y1.copy()
        
    return(X0,Y0)




def calc_correlation(matrix, D, C, key, variant, set_used):
    set_used_copy=set_used
    x=[]
    y=[]
    
    if set_used=='up':
        set_used=D['Test']['Up-regulated']
    elif set_used=='down':
        set_used=D['Test']['Down-regulated']
    elif set_used=='up+down':
        set_used=[*D['Test']['Up-regulated'], *D['Test']['Down-regulated']]
    elif set_used=='unknown':
        set_used=D['Test']['Unknown']
    elif set_used=='not':
        set_used=D['Test']['Not regulated']
    elif set_used=='empty':
        set_used=D['Test']['Empty']
    elif set_used=='up+down+not':
        set_used=[*D['Test']['Up-regulated'], *D['Test']['Down-regulated'], *D['Test']['Not regulated']]
    elif set_used=='up+down+not+unknown':
        set_used=[*D['Test']['Up-regulated'], *D['Test']['Down-regulated'], *D['Test']['Not regulated'], *D['Test']['Unknown']]
    else:
        print(f'Error: set_used {key} is not in [up, down, not, unknown, empty, up+down, up+down+not, up+down+not+unknown]')
    
    
    if 'empty' in set_used_copy:
        for i in set_used:
            y.append(C[i])
            
        sns.histplot(y)
        return((-999,-999),(-999,-999))
    else:
        for i in set_used:
            if key in ['pEC50', 'log(Fold_change_predicted)']:
                x.append(float(matrix.layers[key][i]))
            else:
                x.append(np.log(float(matrix.layers[key][i])))
            y.append(C[i])

        pearson=scipy.stats.pearsonr(x,y)
        spearman=scipy.stats.spearmanr(x,y)
        print(f'Spearman correlation: {spearman[0]}, p-value: {spearman[1]}')
        print(f'Pearson correlation: {pearson[0]}, p-value: {pearson[1]}')

        sns.kdeplot(x=x[:10**4], y=y[:10**4], fill=True)

        return(pearson, spearman)


def determine_cutoff(df_precission_recall, error, method):

    l=-1
    for i in range(len(df_precission_recall)-1):
        if 1-error<df_precission_recall[method][i+1] and 1-error>df_precission_recall[method][i]:
            l=i
            break

    if l==-1:
        print(f'It is not possible to chose Recall={1-error}')
    else:
        # Linear interpolation to find the cutoff

        c0=df_precission_recall['Cutoff'][i]
        c1=df_precission_recall['Cutoff'][i+1]

        p0=df_precission_recall[method][i]
        p1=df_precission_recall[method][i+1]

        cut=c0+(c1-c0)*(1-error-p0)/(p1-p0)

    return(cut)

def determine_cutoffs(df_precission_recall, method='Precission'):
    Dict={}
    for error in [0.5,0.1,0.05,0.01,0.001]:
        if method=='Precission':
            Dict[f'Type 1 error: {error}']=determine_cutoff(df_precission_recall, error=error, method=method)
        elif method=='Recall':
            Dict[f'Type 2 error: {error}']=determine_cutoff(df_precission_recall, error=error, method=method)
        else:
            print('ERROR: method {method} is not in [Precission, Recall]')
    return(Dict)




def fast_error_to_cutoff(df_not_reg, df_reg, cut):
    false_positives=np.array(df_not_reg['Score']>cut).sum()
    true_positives=np.array(df_reg['Score']>=cut).sum()
    precission=true_positives/(true_positives+false_positives)
    return(precission)


def score_to_confidence(matrix, df):
    l0=len(matrix.obs)
    l1=len(matrix.var)
    
    df_not_reg=df[df['Set']=='Not regulated']
    variant=df[df['Set']!='Not regulated']['Set'].iloc[0].split('(')[1][:-1]
    
    df_reg=df[df['Set']==f'Regulated ({variant})']
    
    C=matrix.layers['ALS_scores']
    
    Confidence=np.zeros((l0,l1))
    for i in range(l0):
        for j in range(l1):
            Confidence[i,j]=fast_error_to_cutoff(df_not_reg, df_reg, C[i,j])
    matrix.layers['Confidence']=Confidence
    return(matrix)



def add_ALS_embedding(matrix, model):
    # Adds the embedding of the drugs and the peptides
    matrix.varm['Drug_embedding']=gpu_matrix_to_numpy(model.user_factors)
    matrix.obsm['Peptide_embedding']=gpu_matrix_to_numpy(model.item_factors)
    return(matrix)


import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import scanpy as sc
import scvelo as scv
import plotly.express as px

def colors_based_on_phylogeny(matrix, Path):
    import matplotlib.colors as colors

    Drug_colors=np.load(f'{Path}/Drug_classes_colors.npy', allow_pickle=True)
    Drug_colors=dict(enumerate(Drug_colors.flatten()))[0]

    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_2d_umap'], obs=matrix.var)
    Emb_matrix.obsm['X_umap']=matrix.varm['Drug_2d_umap']
    Emb_matrix.obs['Drug Class combined']=Emb_matrix.obs['Drug Class combined'].astype('category')


    col=[]
    for i in Emb_matrix.obs['Drug Class combined'].cat.categories:
        col.append(Drug_colors[i])

    col2=[]
    for i in Emb_matrix.obs['Drug Class combined']:
        if not colors.rgb2hex(Drug_colors[i]) in col2:
            col2.append(colors.rgb2hex(Drug_colors[i]))
    return(col, col2)





def calc_umaps(matrix, neighbors_drugs=5, neighbors_peptides=15):
    # 2d drug umap
    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_embedding'], obs=matrix.var)
    sc.pp.neighbors(Emb_matrix, n_neighbors=neighbors_drugs)
    sc.tl.umap(Emb_matrix)
    matrix.varm['Drug_2d_umap']=Emb_matrix.obsm['X_umap']
    matrix.uns['Drug_emb_2d']=Emb_matrix
    # 3d drug umap
    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_embedding'], obs=matrix.var)
    sc.pp.neighbors(Emb_matrix, n_neighbors=neighbors_drugs)
    sc.tl.umap(Emb_matrix, n_components=3)
    matrix.varm['Drug_3d_umap']=Emb_matrix.obsm['X_umap']
    matrix.uns['Drug_emb_3d']=Emb_matrix

    # 2d peptide umap
    Emb_matrix=anndata.AnnData(X=matrix.obsm['Peptide_embedding'], obs=matrix.obs)
    sc.pp.neighbors(Emb_matrix, n_neighbors=neighbors_peptides)
    sc.tl.umap(Emb_matrix)
    matrix.obsm['Peptide_2d_umap']=Emb_matrix.obsm['X_umap']
    matrix.uns['Peptide_emb_2d']=Emb_matrix
    # 3d peptide umap
    Emb_matrix=anndata.AnnData(X=matrix.obsm['Peptide_embedding'], obs=matrix.obs)
    sc.pp.neighbors(Emb_matrix, n_neighbors=neighbors_peptides)
    sc.tl.umap(Emb_matrix, n_components=3)
    matrix.obsm['Peptide_3d_umap']=Emb_matrix.obsm['X_umap']
    matrix.uns['Peptide_emb_3d']=Emb_matrix
    
    return(matrix)

def plot_drug_umap(matrix, Path, color_key='Drug Class combined', interactive=True, palette='Phylogeny'):
    sns.set_style("whitegrid")
    if palette=='Phylogeny':
        col, palette_chosen=colors_based_on_phylogeny(matrix, Path)
    else:
        palette_chosen=px.colors.qualitative.Light24
        col=px.colors.qualitative.Light24

    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_2d_umap'], obs=matrix.var)
    Emb_matrix.obsm['X_umap']=matrix.varm['Drug_2d_umap']
    
    if interactive:
        M=matrix.varm['Drug_2d_umap']
        (x,y)=(M[:,0], M[:,1])


        dfs=pd.DataFrame({'x':x, 'y':y, color_key:matrix.var[color_key], 'Drug Names':matrix.var['Drug Names'],
                         'Drug Class':matrix.var['Drug Class']})

        fig = px.scatter(dfs, x='x', y='y', color=color_key, color_discrete_sequence=palette_chosen,
                               custom_data=[color_key, 'Drug Names', 'Drug Class'])
        fig.update_traces(
            hovertemplate="<br>".join([
                "x: %{x}",
                "y: %{y}",
                "color_key: %{customdata[0]}",
                "Names: %{customdata[1]}",
                "Drug Class: %{customdata[2]}"
            ]))

        fig.update_traces(marker=dict(size=20))

        fig.show()
    else:
        scv.pl.scatter(Emb_matrix, color=color_key, basis='umap', legend_loc='right', palette=col)
        
    
    
def plot_kinobead_umap(matrix, color_key='FGFR1', interactive=True):
    import plotly.express as px
    df=matrix.uns['Drug_class_matrix'].copy()
    df2=matrix.uns['Kinobead_matrix'].copy()
    kinobead_affinity=[a for a in df2[color_key]]
    kinobead_affinity=[0 if np.isnan(a) else a for a in kinobead_affinity]
    log_kinobead_affinity=[0 if a==0 else np.log(a) for a in kinobead_affinity]

    Emb_matrix=anndata.AnnData(X=matrix.varm['Drug_2d_umap'], obs=matrix.var)
    Emb_matrix.obsm['X_umap']=matrix.varm['Drug_2d_umap']
    Emb_matrix.obs[color_key]=log_kinobead_affinity
    
    
    if interactive:

        M=matrix.varm['Drug_2d_umap']
        (x,y)=(M[:,0], M[:,1])


        dfs=pd.DataFrame({'x':x, 'y':y, color_key:log_kinobead_affinity, 'Drug Names':[a for a in matrix.var['Drug Names']],
                         'Drug Class':[a for a in matrix.var['Drug Class']],
                         'Kinobead_affinity':kinobead_affinity})
        fig = px.scatter(dfs, x='x', y='y', color=color_key, color_discrete_sequence=px.colors.qualitative.Light24,
                               custom_data=[color_key, 'Drug Names', 'Drug Class', 'Kinobead_affinity'])
        fig.update_traces(
            hovertemplate="<br>".join([
                "x: %{x}",
                "y: %{y}",
                "color_key: %{customdata[0]}",
                "Names: %{customdata[1]}",
                "Drug Class: %{customdata[2]}",
                "Kinobead_affinity: %{customdata[3]}",
            ]))
        fig.update_traces(marker=dict(size=20))

        fig.show()

    else:
        scv.pl.scatter(Emb_matrix, color=color_key, basis='umap', legend_loc='right')
    
    
def plot_peptide_umap(matrix, color_key='RSRC2', interactive=True):
    import plotly.express as px

    Emb_matrix=anndata.AnnData(X=matrix.obsm['Peptide_2d_umap'], obs=matrix.obs)
    Emb_matrix.obsm['X_umap']=matrix.obsm['Peptide_2d_umap']
    
    df=matrix.uns['KEGG_matrix'].copy()
    Emb_matrix.obs[color_key]=np.array(matrix.obs['Gene Names']==color_key).astype(int)
    
    if interactive:
        M=matrix.obsm['Peptide_2d_umap']
        (x,y)=(M[:,0], M[:,1])

        dfs=pd.DataFrame({'x':x, 'y':y, 
                          color_key:np.array(matrix.obs['Gene Names']==color_key).astype(int), 
                          'Peptide': [a for a in matrix.obs['Modified sequence']],
                          'Gene': [a for a in matrix.obs['Gene Names']],
                          'Reference protein': [a for a in matrix.obs['Reference_protein']],
                          'Ph_Site(s)': [a for a in matrix.obs['Ph_site']]})
        dfs=dfs.sort_values(color_key)

        fig = px.scatter(dfs, x='x', y='y', color=color_key+' '*10,
                        custom_data=[color_key, 'Peptide', 'Gene', 'Reference protein', 'Ph_Site(s)'])

        fig.update_traces(
            hovertemplate="<br>".join([
                "x: %{x}",
                "y: %{y}",
                "color_key: %{customdata[0]}",
                "Peptide: %{customdata[1]}",
                "Gene: %{customdata[2]}",
                "Reference protein: %{customdata[3]}",
                "Ph_Site(s): %{customdata[4]}",
            ]))

        fig.update_traces(marker=dict(size=7))

        fig.show()
    else:
        scv.pl.scatter(Emb_matrix, color=color_key, basis='umap', legend_loc='right', size=100)
    
    
def plot_KEGG_umap(matrix, color_key='Apelin signaling pathway', interactive=True):
    import plotly.express as px

    Emb_matrix=anndata.AnnData(X=matrix.obsm['Peptide_2d_umap'], obs=matrix.obs)
    Emb_matrix.obsm['X_umap']=matrix.obsm['Peptide_2d_umap']
    
    df=matrix.uns['KEGG_matrix'].copy()
    Emb_matrix.obs[color_key]=[a for a in df[color_key]]


    if interactive:    
        M=matrix.obsm['Peptide_2d_umap']
        (x,y)=(M[:,0], M[:,1])

        dfs=pd.DataFrame({'x':x, 'y':y, color_key:Emb_matrix.obs[color_key], 
                          'Peptide': [a for a in matrix.obs['Modified sequence']],
                          'Gene': [a for a in matrix.obs['Gene Names']],
                          'Reference protein': [a for a in matrix.obs['Reference_protein']],
                          'Ph_Site(s)': [a for a in matrix.obs['Ph_site']]})
        dfs=dfs.sort_values(color_key)

        fig = px.scatter(dfs, x='x', y='y', color=color_key,
                        custom_data=[color_key, 'Peptide', 'Gene', 'Reference protein', 'Ph_Site(s)'])

        fig.update_traces(
            hovertemplate="<br>".join([
                "x: %{x}",
                "y: %{y}",
                "color_key: %{customdata[0]}",
                "Peptide: %{customdata[1]}",
                "Gene: %{customdata[2]}",
                "Reference protein: %{customdata[3]}",
                "Ph_Site(s): %{customdata[4]}",
            ]))

        fig.update_traces(marker=dict(size=7))

        fig.show()
    else:
        scv.pl.scatter(Emb_matrix, color=color_key, basis='umap', legend_loc='right', size=100)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['psite_recommender']

package_data = \
{'': ['*']}

install_requires = \
['anndata>=0.8.0,<0.9.0',
 'implicit>=0.6.1,<0.7.0',
 'numpy>=1.23.4,<2.0.0',
 'pandas>=1.5.1,<2.0.0',
 'scanpy>=1.9.1,<2.0.0',
 'scvelo>=0.2.4,<0.3.0',
 'seaborn>=0.12.1,<0.13.0',
 'statsmodels>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'psite-recommender',
    'version': '0.1.0',
    'description': 'RS for Proteomics',
    'long_description': '# Recommender_System_Proteomics\nScripts to run Recommender Systems on Proteomics data.\n\n## Downloading:\n  - click on the green "Code" button on the top right and Download ZIP, then unzip it  \n  or use  \n  - git clone https://github.com/ManuelGander/Recommender_System_Proteomics.git\n  \n## Running the scripts:\n  - The notebook "Full_script.ipynb" contains all the information needed to run the scripts\n',
    'author': 'Manuel Gander',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import pandas as pd
import numpy as np
import seaborn as sns
import anndata
import matplotlib.pyplot as plt
import scipy
from sklearn import metrics, preprocessing
import tensorflow as tf
## for deep learning
from tensorflow.keras import models, layers, utils  #(2.6.0)


### To restrict memory used
gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
    try:
        tf.config.experimental.set_virtual_device_configuration(gpus[0],[tf.config.experimental.VirtualDeviceConfiguration(memory_limit=1024)])
    except RuntimeError as e:
        print(e)

    

def matrix_factorization(matrix, dim0=30):
    n0,n1=matrix.shape

    # Users (1,embedding_size)
    xusers_in = layers.Input(name="xusers_in", shape=(1,))
    xusers_emb = layers.Embedding(name="xusers_emb", input_dim=n0, output_dim=dim0)(xusers_in)
    xusers = layers.Reshape(name='xusers', target_shape=(dim0,))(xusers_emb)
    # Products (1,embedding_size)
    xproducts_in = layers.Input(name="xproducts_in", shape=(1,))
    xproducts_emb = layers.Embedding(name="xproducts_emb", input_dim=n1, output_dim=dim0)(xproducts_in)
    xproducts = layers.Reshape(name='xproducts', target_shape=(dim0,))(xproducts_emb)
    # Product (1)
    xx = layers.Dot(name='xx', normalize=True, axes=1)([xusers, xproducts])
    # Predict ratings (1)
    y_out = layers.Dense(name="y_out", units=1, activation='linear')(xx)
    # Compile
    model = models.Model(inputs=[xusers_in,xproducts_in], outputs=y_out, name="CollaborativeFiltering")
    model.compile(optimizer='adam', loss='mean_absolute_error', metrics=['mean_absolute_percentage_error'])

    return(model)


def matrix_factorization_and_depp_learning(matrix, dim0=30, dim1=30, dim_dl=3):
    n0,n1=matrix.shape


    # Input layer
    xusers_in = layers.Input(name="xusers_in", shape=(1,))
    xproducts_in = layers.Input(name="xproducts_in", shape=(1,))

    # A) Matrix Factorization
    ## embeddings and reshape
    cf_xusers_emb = layers.Embedding(name="cf_xusers_emb", input_dim=n0, output_dim=dim0)(xusers_in)
    cf_xusers = layers.Reshape(name='cf_xusers', target_shape=(dim0,))(cf_xusers_emb)
    ## embeddings and reshape
    cf_xproducts_emb = layers.Embedding(name="cf_xproducts_emb", input_dim=n1, output_dim=dim0)(xproducts_in)
    cf_xproducts = layers.Reshape(name='cf_xproducts', target_shape=(dim0,))(cf_xproducts_emb)
    ## product
    cf_xx = layers.Dot(name='cf_xx', normalize=True, axes=1)([cf_xusers, cf_xproducts])



    # B) Neural Network
    ## embeddings and reshape
    nn_xusers_emb = layers.Embedding(name="nn_xusers_emb", input_dim=n0, output_dim=dim1)(xusers_in)
    nn_xusers = layers.Reshape(name='nn_xusers', target_shape=(dim1,))(nn_xusers_emb)
    ## embeddings and reshape
    nn_xproducts_emb = layers.Embedding(name="nn_xproducts_emb", input_dim=n1, output_dim=dim1)(xproducts_in)
    nn_xproducts = layers.Reshape(name='nn_xproducts', target_shape=(dim1,))(nn_xproducts_emb)
    ## concat and dense
    nn_xx = layers.Concatenate()([nn_xusers, nn_xproducts])
    nn_xx = layers.Dense(name="nn_xx", units=int(dim_dl), activation='relu')(nn_xx)

    # Merge A & B
    y_out = layers.Concatenate()([cf_xx, nn_xx])
    y_out = layers.Dense(name="y_out", units=1, activation='linear')(y_out)
    # Compile
    model = models.Model(inputs=[xusers_in,xproducts_in], outputs=y_out, name="Neural_CollaborativeFiltering")
    model.compile(optimizer='adam', loss='mean_absolute_error', metrics=['mean_absolute_percentage_error'])
    
    return(model)


def mf_and_dl_and_additional_features(matrix, dim0=5, dim1=5, dim_dl=2, dim_features=5, dim_dl_and_features=2):
    n0,n1=matrix.shape


    ################### COLLABORATIVE FILTERING ########################
    # Input layer
    xusers_in = layers.Input(name="xusers_in", shape=(1,))
    xproducts_in = layers.Input(name="xproducts_in", shape=(1,))
    # A) Matrix Factorization
    ## embeddings and reshape
    cf_xusers_emb = layers.Embedding(name="cf_xusers_emb", input_dim=n0, output_dim=dim0)(xusers_in)
    cf_xusers = layers.Reshape(name='cf_xusers', target_shape=(dim0,))(cf_xusers_emb)
    ## embeddings and reshape
    cf_xproducts_emb = layers.Embedding(name="cf_xproducts_emb", input_dim=n1, output_dim=dim0)(xproducts_in)
    cf_xproducts = layers.Reshape(name='cf_xproducts', target_shape=(dim0,))(cf_xproducts_emb)
    ## product
    cf_xx = layers.Dot(name='cf_xx', normalize=True, axes=1)([cf_xusers, cf_xproducts])
    # B) Neural Network
    ## embeddings and reshape
    nn_xusers_emb = layers.Embedding(name="nn_xusers_emb", input_dim=n0, output_dim=dim1)(xusers_in)
    nn_xusers = layers.Reshape(name='nn_xusers', target_shape=(dim1,))(nn_xusers_emb)
    ## embeddings and reshape
    nn_xproducts_emb = layers.Embedding(name="nn_xproducts_emb", input_dim=n1, output_dim=dim1)(xproducts_in)
    nn_xproducts = layers.Reshape(name='nn_xproducts', target_shape=(dim1,))(nn_xproducts_emb)
    ## concat and dense
    nn_xx = layers.Concatenate()([nn_xusers, nn_xproducts])
    nn_xx = layers.Dense(name="nn_xx", units=dim_dl, activation='relu')(nn_xx)


    ######################### CONTENT BASED ############################
    pEC50 = layers.Input(name="pEC50", shape=(1,))
    features0 = layers.Dense(name="features0", units=1, activation='relu')(pEC50)

    SN = layers.Input(name="SN", shape=(1,))
    features1 = layers.Dense(name="features1", units=1, activation='relu')(SN)

    FC = layers.Input(name="FC", shape=(1,))
    features2 = layers.Dense(name="features2", units=1, activation='relu')(FC)

    ######################  Merge additional information     #######################################
    # Merge all
    merge_features = layers.Concatenate()([features0, features1, features2])
    merge_features = layers.Dense(name="merge_features", units=dim_features, activation='relu')(merge_features)


    ######################  Merge additional information     #######################################
    # Merge all
    merger = layers.Concatenate()([nn_xx, merge_features])
    merger = layers.Dense(name="merger", units=dim_dl_and_features, activation='relu')(merger)


    ########################## OUTPUT ##################################
    # Merge all
    
    y_out = layers.Concatenate()([cf_xx, merger])
    y_out = layers.Dense(name="y_out", units=1, activation='linear')(y_out)
    # Compile
    model = models.Model(inputs=[xusers_in, xproducts_in, pEC50, SN, FC], outputs=y_out, name="Hybrid_Model")
    model.compile(optimizer='adam', loss='mean_absolute_error', metrics=['mean_absolute_percentage_error'])
    return(model)




def get_additional_information(matrix, in0, in1):
    
    # Get them all log_transformed
    pEC50=[]
    SN=[]
    FC=[]
    
    for l in range(len(in0)):
        pEC50.append(float(matrix.layers['pEC50'][in0[l],in1[l]]))
        SN.append(np.log(float(matrix.layers['Total SN'][in0[l],in1[l]])))
        FC.append(float(matrix.layers['log(Fold_change_predicted)'][in0[l],in1[l]]))
    
    pEC50=np.array(pEC50)
    SN=np.array(SN)
    FC=np.array(FC)
    
    return(pEC50, SN, FC)


def model_to_output_test_set(matrix, D, model, test_only=False):
    n0,n1=matrix.shape
    C=np.zeros((n0,n1))
    ind0=[]
    ind1=[]
    
    if test_only:
        Dt=D['Test']
        for i in [*Dt['Up-regulated'], *Dt['Down-regulated'],  *Dt['Not regulated']]:
            ind0.append(i[0])
            ind1.append(i[1])

    else:
        for i in range(n0):
            for j in range(n1):
                if matrix.X[i,j]!='Empty':
                    ind0.append(i)
                    ind1.append(j)
                    
    ind0=np.array(ind0)
    ind1=np.array(ind1)
    
    EC50, SN, FC=get_additional_information(matrix, ind0, ind1)
    
    res= model.predict([ind0, ind1, EC50, SN, FC]).T[0]
    
    l=0
    for i in range(len(ind0)):
        C[ind0[i], ind1[i]]= res[i]
    return(C)



def model_to_output(matrix, D, model, test_only=False):
    n0,n1=matrix.shape
    C=np.zeros((n0,n1))
    
    ind0=[]
    ind1=[]
    
    if test_only:
        Dt=D['Test']
        for i in [*Dt['Up-regulated'], *Dt['Down-regulated'],  *Dt['Not regulated']]:
            ind0.append(i[0])
            ind1.append(i[1])

    else:
        for i in range(n0):
            for j in range(n1):
                ind0.append(i)
                ind1.append(j)
                
    ind0=np.array(ind0)
    ind1=np.array(ind1)
    
    res= model.predict([ind0, ind1]).T[0]
    
    for i in range(len(ind0)):
        C[ind0[i],ind1[i]]= res[i]
    return(C)